"""
Test cases for cli2.Node
"""
from cli2 import Node


def example_function(*args, **kwargs):
    """
    Example function docstring where the first sentence unfortunnately spreads
    over the next line.
    """
    return f"""
    You have called example_function with:

    args={args}
    kwargs={kwargs}
    """


example_function.cli2 = dict(color='pink')


class ExampleClass:
    def example_method(self): pass


example_object = ExampleClass()


class ExampleClassCallable:
    def __call__(self, *args, **kwargs): return (args, kwargs)


example_object_callable = ExampleClassCallable()


def test_eq():
    assert Node.factory('cli2') == Node.factory('cli2')


def test_module():
    from cli2 import test_node
    node = Node.factory('cli2.test_node')
    assert node.target == test_node
    assert node.type == 'module'
    assert str(node) == 'cli2.test_node'
    assert repr(node) == 'Node(cli2.test_node)'
    assert example_object_callable in node.callables

    node = Node('test_node', test_node)
    assert node.target == test_node
    assert node.type == 'module'


example_list = [lambda: True]


def test_list():
    node = Node.factory('cli2.test_node.example_list.0')
    assert node.target == example_list[0]


example_dict = dict(a=lambda: True)


def test_dict():
    node = Node.factory('cli2.test_node.example_dict.a')
    assert node.target == example_dict['a']


def test_unknown():
    node = Node.factory('lollololololololollolool')
    assert not node.module
    assert not node.target


def test_function():
    node = Node.factory('cli2.test_node.example_function')
    assert node.target == example_function
    assert node.type == 'function'
    assert not node.callables


def test_class():
    node = Node.factory('cli2.test_node.ExampleClass')
    assert node.target == ExampleClass
    assert Node('', ExampleClass.example_method) in node.callables


def test_object():
    node = Node.factory('cli2.test_node.example_object')
    assert node.target == example_object
    assert example_object.example_method in node.callables


def test_callable_class():
    node = Node.factory('cli2.test_node.ExampleClassCallable')
    assert node.target == ExampleClassCallable


def test_callable_object():
    node = Node.factory('cli2.test_node.example_object_callable')
    assert node.target == example_object_callable



from .group import Group


def test_group_command_not_found():
    assert 'Command a not found' in Group()(['a', 'b'])


def test_group_subcommand_not_found():
    group = Group()
    group['a'] = Group(name='a')
    assert 'Command b not found' in group(['a', 'b'])


def test_group_no_command():
    assert 'No command' in Group()([])


def test_missing_arg():
    cmd = Group().cmd(lambda b: True, name='a')
    assert "missing 1 required positional argument: 'b'" in cmd(['a'])


def test_repr():
    assert repr(Group('foo')) == 'Group(foo)'


def test_help():
    def foo():
        """foodoc"""
    group = Group('lol', doc='loldoc')
    group.cmd(foo)
    assert 'foodoc' in group([])
    assert 'loldoc' in group([])
    assert 'foodoc' in group(['help'])
    assert 'foodoc' in group(['help', 'foo'])
    assert 'not found' in group(['help', 'lol'])


def test_load_module():
    from cli2 import test_group
    group = Group()
    group.load(test_group)
    assert 'test_load_module' in group


def test_load_object():
    class Lol:
        def __call__(self): pass  # noqa
    group = Group()
    group.load(Lol())
    assert 'Lol' in group



import inspect
import importlib
import types


class Node:
    def __init__(self, name, target, module=None):
        self.name = name
        self.target = target

        if module:
            self.module = module
        elif isinstance(target, types.ModuleType):
            self.module = target
        elif hasattr(target, '__module__'):
            self.module = target.__module__
        else:
            self.module = None

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.target == other.target
        return self.target == other

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'Node({self.name})'

    @property
    def type(self):
        if isinstance(self.target, types.ModuleType):
            return 'module'
        if isinstance(self.target, types.FunctionType):
            return 'function'

    @property
    def callables(self):
        results = []
        builtins = (types.BuiltinFunctionType, types.BuiltinMethodType)
        for name, member in inspect.getmembers(self.target):
            if not callable(member):
                continue
            if name.startswith('__') or isinstance(member, builtins):
                continue  # skip builtins
            results.append(Node(name, member))
        return results

    @classmethod
    def factory(cls, name):
        module = None
        parts = name.split('.')
        for i, part in reversed(list(enumerate(parts))):
            modname = '.'.join(parts[:i + 1])

            if not modname:
                break

            try:
                module = importlib.import_module(modname)
            except ImportError:
                continue
            else:
                break

        if module:
            ret = module
            for part in parts[i + 1:]:
                if isinstance(ret, dict) and part in ret:
                    ret = ret.get(part)
                elif isinstance(ret, list) and part.isnumeric():
                    ret = ret[int(part)]
                else:
                    ret = getattr(ret, part, None)
        else:
            ret = None

        return cls(name, ret, module)



from cli2.cli import main


def test_call():
    result = main(['cli2.test_node.example_function', 'x', 'y=z'])
    assert "args=('x',)" in result
    assert "kwargs={'y': 'z'}" in result


def test_doc():
    result = main(['cli2.test_node'])
    assert 'example_function' in result

    result = main(['help', 'cli2.test_node'])
    assert 'example_function' in result

    result = main([])
    assert 'help' in result



import inspect
import subprocess
import sys

from .colors import colors
from .command import Command


def termsize():
    try:
        rows, columns = subprocess.check_output(['stty', 'size']).split()
    except subprocess.CalledProcessError:
        return 180, 80
    return int(rows), int(columns)


class Group(dict):
    def __init__(self, name=None, doc=None, color=None):
        self.name = name
        self.doc = doc or inspect.getdoc(self)
        self.color = color or colors.green

    def cmd(self, target, name=None):
        cmd = Command(target, name)
        self[cmd.name] = cmd
        return self

    def help(self, error=None, short=False):
        output = []
        if error:
            output.append(error + '\n')

        if self.doc:
            output.append(self.doc + '\n')

        if not len(self):
            return '\n'.join(output)

        namewidth = 2 + max([len(key) for key in self]) + 2
        descwidth = termsize()[1] - namewidth

        for name, command in self.items():
            doc = command.help(short=True)
            out = ''
            for char in doc:
                if len(out) == descwidth and '\n' not in out:
                    out += '\n' + ' ' * (namewidth)
                out += char

            line = [
                '  ',
                command.color,
                name,
                colors.reset,
                ' ' * (namewidth - len(str(name)) - 4),
                '  ',
                out,
            ]
            output.append(''.join(map(str, line)))
        return '\n'.join(output)

    def load(self, obj):
        for name in dir(obj):
            if name == '__call__':
                target = obj
            elif name.startswith('__'):
                continue
            else:
                target = getattr(obj, name)

            if callable(target):
                cmd = Command(target)
                self[cmd.name] = cmd
                continue
            else:
                self[name] = Group(name)

    def __call__(self, argv=None):
        argv = argv if argv is not None else sys.argv[1:]
        if not argv:
            return self.help('No command provided, showing help.')

        if argv[0] in self:
            result = self[argv[0]](argv[1:])
        elif argv[0] == 'help':
            if len(argv) > 1:
                if argv[1] in self:
                    return self[argv[1]].help()
                else:
                    return self.help(
                        f'Command {argv[0]} not found, showing help.'
                    )
            else:
                return self.help()
        else:
            return self.help(f'Command {argv[0]} not found, showing help.')

        return result

    def __repr__(self):
        return f'Group({self.name})'



import asyncio
import inspect
import sys

from .argument import Argument
from .colors import colors


class Command(dict):
    def __init__(self, target, name=None, doc=None, color=None):
        self.target = target
        self.name = name or getattr(target, '__name__', type(target).__name__)
        self.doc = doc or inspect.getdoc(target)
        self.color = color or colors.orange

        overrides = getattr(target, 'cli2', {})
        for key, value in overrides.items():
            setattr(self, key, value)

        if self.color in colors.__dict__:
            self.color = getattr(colors, self.color)

        self.sig = inspect.signature(target)

        for name, param in self.sig.parameters.items():
            self[name] = Argument(self, param)
            overrides = getattr(self.target, 'cli2_' + name, {})
            for key, value in overrides.items():
                setattr(self[name], key, value)

    def help(self, error=None, short=False):
        output = []

        if error:
            output.append(error + '\n')

        if self.doc:
            if short:
                # get the first sentence
                sentence = ''
                for char in self.doc.replace('\n', ' '):
                    if char == '.':
                        break
                    sentence += char
                output.append(sentence)
            else:
                output.append(self.doc + '\n')

        return '\n'.join(output)

    def parse(self, *argv):
        self.bound = self.sig.bind_partial()
        extra = []
        for current in argv:
            taken = False
            for arg in self.values():
                if not arg.accepts:
                    continue

                if arg.iskw:
                    # we have reached keyword argument sequence
                    # see if another, further arg matches per alias and takes
                    reached = False
                    for _arg in self.values():
                        if taken:
                            break
                        if _arg == arg:
                            reached = True
                            continue
                        if not reached:
                            continue
                        if _arg.aliasmatch(current):
                            taken = _arg.take(current)

                if taken:
                    break

                taken = arg.take(current)
                if taken:
                    break

            if not taken:
                extra.append(current)

        if extra:
            return 'No parameters for these arguments: ' + ', '.join(extra)

    def __call__(self, argv=None):
        error = self.parse(*(argv if argv is not None else sys.argv[1:]))
        if error:
            return self.help(error)

        try:
            result = self.target(*self.bound.args, **self.bound.kwargs)
        except TypeError as exc:
            rep = getattr(self.target, '__name__')
            error = str(exc)
            if error.startswith(rep):
                return self.help(error.replace(rep, self.name))
            raise

        if inspect.iscoroutine(result):
            result = asyncio.run(result)
        return result



import pytest

from .command import Command


def test_int():
    def foo(one: int): return one
    cmd = Command(foo)
    cmd.parse('1')
    assert cmd['one'].value == 1
    assert not cmd['one'].accepts
    assert cmd(['1']) == 1
    assert repr(cmd['one']) == 'one'


def test_vararg():
    def foo(*one): return one
    cmd = Command(foo)
    cmd.parse('a', 'b')
    assert cmd['one'].value == ['a', 'b']
    assert cmd['one'].accepts
    assert cmd(['a', 'b']) == ('a', 'b')


def test_kwarg():
    def foo(one=None): return one
    cmd = Command(foo)
    cmd.parse('one=b')
    assert cmd['one'].value == 'b'
    assert not cmd['one'].accepts
    assert cmd(['one=b']) == 'b'


def test_varkwarg():
    def foo(**one): return one
    cmd = Command(foo)
    cmd.parse('a=b', 'c=d')
    assert cmd['one'].value == dict(a='b', c='d')
    assert cmd['one'].accepts


def test_skip():
    def foo(a=None, b=None, c=None):
        return (a, b, c)
    cmd = Command(foo)
    assert cmd(['b=x']) == (None, 'x', None)


def test_nested_typeerror():
    # Test that TypeError unrelated to top level function call are not
    # swallowed to display help
    def foo():
        raise TypeError('Lol')
    cmd = Command(foo)
    with pytest.raises(TypeError):
        cmd([])


def test_vararg_varkwarg_natural():
    def foo(*one, **two): return (one, two)
    cmd = Command(foo)
    cmd.parse('x', 'y', 'a=b', 'c=d')
    assert cmd['one'].value == ['x', 'y']
    assert cmd['two'].value == dict(a='b', c='d')
    assert cmd(['x', 'y', 'a=b', 'c=d']) == (('x', 'y'), dict(a='b', c='d'))


def test_vararg_varkwarg_asterisk():
    def foo(*one, **two): return (one, two)
    cmd = Command(foo)
    cmd.parse('*["x"]', '**{"y" :"z"}')
    assert cmd['one'].value == ['x']
    assert cmd['two'].value == dict(y='z')


def test_bool():
    def foo(one: bool): return one
    cmd = Command(foo)

    cmd.parse('yes')
    assert cmd['one'].value is True
    assert not cmd['one'].accepts
    assert cmd(['yes']) is True

    for i in ('0', 'no', 'False'):
        cmd.parse(i)
        assert cmd['one'].value is False
        assert not cmd['one'].accepts
        assert cmd([i]) is False


def test_bool_flag():
    def foo(one: bool): return one
    foo.cli2_one = dict(alias='-o')
    cmd = Command(foo)
    cmd.parse('-o')
    assert cmd['one'].value is True


def test_bool_flag_negate():
    def foo(one: bool): return one
    foo.cli2_one = dict(alias='-o', negate='!o')
    cmd = Command(foo)
    cmd.parse('!o')
    assert cmd['one'].value is False


def test_json_cast():
    import json
    def foo(one): return one
    foo.cli2_one = dict(cast=lambda v: json.loads(v))
    cmd = Command(foo)
    cmd.parse('[1]')
    assert cmd['one'].value == [1]


def test_further_search():
    def foo(a=None, one: bool = None): return one
    cmd = Command(foo)
    assert cmd['one'].match('one=yes') == 'yes'

    cmd.parse('one=yes')
    assert cmd['one'].value is True
    assert not cmd['one'].accepts
    assert cmd(['one=yes']) is True


def test_list():
    def foo(one: list = None): return one
    cmd = Command(foo)
    assert cmd['one'].match('one=[1]') == '[1]'

    cmd.parse('one=[1]')
    assert cmd['one'].value == [1]
    assert not cmd['one'].accepts

    assert cmd(['one=[1]']) == [1]
    assert cmd(['[1]']) == [1]

    # simple syntax for simple list of strings
    assert cmd(['one=a,b']) == ['a', 'b']


def test_dict():
    def foo(one: dict = None): return one
    cmd = Command(foo)
    assert cmd['one'].match('one={"a": 1}') == '{"a": 1}'

    cmd.parse('one={"a": 1}')
    assert cmd['one'].value == {"a": 1}
    assert not cmd['one'].accepts

    assert cmd(['one={"a": 1}']) == {"a": 1}
    assert cmd(['{"a": 1}']) == {"a": 1}

    # simple syntax for simple dict of strings
    assert cmd(['one=a:b,c:d']) == {"a": "b", "c": "d"}


def test_override():
    def foo(one): return one
    foo.cli2 = dict(color=1, name='lol', doc='foodoc')
    cmd = Command(foo)
    assert cmd.color == 1
    assert cmd.name == 'lol'
    assert cmd.doc == 'foodoc'


def test_cast_override():
    def foo(one): return one
    foo.cli2_one = dict(cast=lambda v: [int(i) for i in v.split(',')])
    cmd = Command(foo)
    cmd.parse('1,2')
    assert cmd['one'].value == [1, 2]
    assert cmd(['1,2']) == [1, 2]


def test_weird_pattern():
    # show off our algorithms weakness, while it's still fresh in my head :)
    def foo(a=None, b=None):
        return (a, b)
    cmd = Command(foo)

    cmd.parse('b=x')
    assert cmd['b'].value == 'x'
    assert cmd(['b=x']) == (None, 'x')

    cmd = Command(foo)
    cmd.parse('a=b=x')
    assert cmd['a'].value == 'b=x'
    assert cmd(['a=b=x']) == ('b=x', None)


class Foo:
    def __call__(self, one: int, two: list, three: bool, *vararg, **varkwarg):
        self.one = one
        self.two = two
        self.three = three
        self.vararg = vararg
        self.varkwarg = varkwarg


def test_stress():
    cmd = Command(Foo())
    cmd(['1', '[2]', 'yes', 'var1', 'var2', 'kw1=1', 'kw2=2'])
    assert cmd.target.one == 1
    assert cmd.target.two == [2]
    assert cmd.target.three is True
    assert cmd.target.vararg == ('var1', 'var2')
    assert cmd.target.varkwarg == dict(kw1='1', kw2='2')


def test_missing():
    def foo(missing):
        """docstring"""
    cmd = Command(foo)
    assert 'missing 1 required' in cmd([])
    assert 'docstring' in cmd([])


def test_extra():
    cmd = Command(lambda: True)
    assert 'No parameters for these' in cmd(['a'])


def test_asyncio():
    async def test():
        return 'foo'
    assert Command(test)([]) == 'foo'



"""
Demonstration of a completely dynamic command line !
"""
import inspect
import sys

import cli2


class ConsoleScript(cli2.Group):
    """
    cli2 makes your python callbacks work on CLI too !

    Show doc and callables for module cli2.test_node:

        cli2 cli2.test_node
        cli2 help cli2.test_node  # alternate

    Call cli2.test_node.example_function with args=['x'] and kwargs=dict(y='z')

        cli2 cli2.test_node.example_function x y=z
    """

    def load(self, dotted_path):
        node = cli2.Node.factory(dotted_path)
        if callable(node.target):
            self[dotted_path] = cli2.Command(node.target)
        else:
            self[dotted_path] = cli2.Group(
                dotted_path,
                doc=inspect.getdoc(node.target),
            )
            self[dotted_path].load(node.target)

    def __call__(self, argv=None):
        argv = argv if argv is not None else sys.argv[1:]

        def arghelp(dotted_path):
            """
            Get help and callables for a dotted_path.
            """
            self.load(dotted_path)
            return self[dotted_path].help()

        self['help'] = cli2.Command(arghelp, color=cli2.c.green)

        if not argv:
            return self.help()

        if argv[0] == 'help':
            if argv:
                self.load(argv[1])
        else:
            self.load(argv[0])

        return super().__call__(argv)


main = ConsoleScript()



import re
import json


class Argument:
    def __init__(self, cmd, param):
        self.cmd = cmd
        self.param = param
        self.alias = param.name if self.iskw else None
        self.negate = None

    def __repr__(self):
        return self.param.name

    @property
    def iskw(self):
        return self.param.kind in (
            self.param.KEYWORD_ONLY,
            self.param.POSITIONAL_OR_KEYWORD,
        )

    @property
    def accepts(self):
        return (
            self.param.name not in self.cmd.bound.arguments
            or self.param.kind in (
                self.param.VAR_POSITIONAL,
                self.param.VAR_KEYWORD,
            )
        )

    @property
    def value(self):
        return self.cmd.bound.arguments[self.param.name]

    @value.setter
    def value(self, value):
        if self.param.kind == self.param.VAR_POSITIONAL:
            self.cmd.bound.arguments.setdefault(self.param.name, [])
            self.cmd.bound.arguments[self.param.name].append(value)
        elif self.param.kind == self.param.VAR_KEYWORD:
            self.cmd.bound.arguments.setdefault(self.param.name, {})
            parts = value.split('=')
            value = '='.join(parts[1:])
            self.cmd.bound.arguments[self.param.name][parts[0]] = value
        else:
            self.cmd.bound.arguments[self.param.name] = value

    def cast(self, value):
        if self.param.annotation == int:
            return int(value)
        if self.param.annotation == bool:
            return value.lower() not in ('', '0', 'no', 'false', self.negate)
        if self.param.annotation == list:
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                return [i.strip() for i in value.split(',')]
        if self.param.annotation == dict:
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                results = dict()
                for token in value.split(','):
                    parts = token.split(':')
                    results[parts[0].strip()] = ':'.join(parts[1:]).strip()
                return results
        return value

    def aliasmatch(self, arg):
        if arg == self.negate:
            return True
        if self.iskw and self.param.annotation == bool and arg == self.alias:
            return True
        return self.alias and arg.startswith(self.alias + '=')

    def match(self, arg):
        if self.aliasmatch(arg):
            if self.param.annotation != bool or '=' in arg:
                return arg[len(self.alias) + 1:]
        return arg

    def take(self, arg):
        if self.param.kind == self.param.VAR_KEYWORD:
            if arg.startswith('**{') and arg.endswith('}'):
                self.cmd.bound.arguments[self.param.name] = json.loads(arg[2:])
                return True

        elif self.param.kind == self.param.VAR_POSITIONAL:
            if arg.startswith('*[') and arg.endswith(']'):
                self.cmd.bound.arguments[self.param.name] = json.loads(arg[1:])
                return True

            # edge case vararg+varkwargs
            # priority to varkwargs for word= and **{}
            last = self.cmd[[*self.cmd.keys()][-1]]
            if last.param.kind == self.param.VAR_KEYWORD:
                if re.match('^\\w+=', arg):
                    return
                elif arg.startswith('**{') and arg.endswith('}'):
                    return

        value = self.match(arg)
        if value is not None:
            self.value = self.cast(value)
            return True



from setuptools import setup


setup(
    name='cli2',
    version='2.0.1.dev1',
    setup_requires='setupmeta',
    extras_require=dict(
        test=[
            'freezegun',
            'pytest',
            'pytest-cov',
        ],
    ),
    author='James Pic',
    author_email='jamespic@gmail.com',
    url='https://yourlabs.io/oss/cli2',
    include_package_data=True,
    license='MIT',
    keywords='cli',
    python_requires='>=3',
)
